import json
import os
from datetime import date, timedelta
from pathlib import Path
from unittest.mock import patch

import pytest
import yaml
from click.testing import CliRunner
from google.api_core.exceptions import NotFound

from bigquery_etl.backfill.parse import (
    BACKFILL_FILE,
    DEFAULT_REASON,
    DEFAULT_WATCHER,
    Backfill,
    BackfillStatus,
)
from bigquery_etl.backfill.utils import (
    BACKFILL_DESTINATION_DATASET,
    BACKFILL_DESTINATION_PROJECT,
    get_backfill_backup_table_name,
    get_backfill_file_from_qualified_table_name,
    get_backfill_staging_qualified_table_name,
    get_entries_from_qualified_table_name,
    get_qualified_table_name_to_entries_map_by_project,
    qualified_table_name_matching,
    validate_metadata_workgroups,
)
from bigquery_etl.cli.backfill import (
    complete,
    create,
    info,
    initiate,
    scheduled,
    validate,
)
from bigquery_etl.cli.stage import QUERY_FILE
from bigquery_etl.deploy import FailedDeployException

DEFAULT_STATUS = BackfillStatus.INITIATE
VALID_REASON = "test_reason"
VALID_WATCHER = "test@example.org"
VALID_BACKFILL = Backfill(
    date(2021, 5, 4),
    date(2021, 1, 3),
    date(2021, 5, 3),
    [date(2021, 2, 3)],
    VALID_REASON,
    [VALID_WATCHER],
    DEFAULT_STATUS,
)
BACKFILL_YAML_TEMPLATE = (
    "2021-05-04:\n"
    "  start_date: 2021-01-03\n"
    "  end_date: 2021-05-03\n"
    "  excluded_dates:\n"
    "  - 2021-02-03\n"
    "  reason: test_reason\n"
    "  watchers:\n"
    "  - test@example.org\n"
    "  status: Initiate\n"
)

VALID_WORKGROUP_ACCESS = [
    dict(
        role="roles/bigquery.dataViewer",
        members=["workgroup:mozilla-confidential"],
    )
]

TABLE_METADATA_CONF = {
    "friendly_name": "test",
    "description": "test",
    "owners": ["test@example.org"],
    "workgroup_access": VALID_WORKGROUP_ACCESS,
}

TABLE_METADATA_CONF_EMPTY_WORKGROUP = {
    "friendly_name": "test",
    "description": "test",
    "owners": ["test@example.org"],
    "workgroup_access": [],
}

TABLE_METADATA_CONF_DEPENDS_ON_PAST = {
    "friendly_name": "test",
    "description": "test",
    "owners": ["test@example.org"],
    "workgroup_access": VALID_WORKGROUP_ACCESS,
    "scheduling": {"depends_on_past": True},
}

DATASET_METADATA_CONF = {
    "friendly_name": "test",
    "description": "test",
    "dataset_base_acl": "derived",
    "workgroup_access": VALID_WORKGROUP_ACCESS,
}

DATASET_METADATA_CONF_EMPTY_WORKGROUP = {
    "friendly_name": "test",
    "description": "test",
    "dataset_base_acl": "derived",
    "workgroup_access": [],
    "default_table_workgroup_access": VALID_WORKGROUP_ACCESS,
}

PARTITIONED_TABLE_METADATA = {
    "friendly_name": "test",
    "description": "test",
    "owners": ["test@example.org"],
    "workgroup_access": VALID_WORKGROUP_ACCESS,
    "bigquery": {
        "time_partitioning": {
            "type": "day",
            "field": "submission_date",
            "require_partition_filter": True,
        }
    },
}

DEFAULT_BILLING_PROJECT = "moz-fx-data-backfill-slots"
VALID_BILLING_PROJECT = "moz-fx-data-backfill-1"
INVALID_BILLING_PROJECT = "mozdata"


class TestBackfill:
    @pytest.fixture
    def runner(self):
        return CliRunner()

    def test_create_backfill(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            result = runner.invoke(
                create,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                    "--start_date=2021-03-01",
                ],
            )

            assert result.exit_code == 0
            assert BACKFILL_FILE in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            backfill_file = SQL_DIR + "/" + BACKFILL_FILE
            backfill = Backfill.entries_from_file(backfill_file)[0]

            assert backfill.entry_date == date.today()
            assert backfill.start_date == date(2021, 3, 1)
            assert backfill.end_date == date.today()
            assert backfill.watchers == [DEFAULT_WATCHER]
            assert backfill.reason == DEFAULT_REASON
            assert backfill.status == DEFAULT_STATUS
            assert backfill.billing_project is None

    def test_create_backfill_with_billing_project(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            result = runner.invoke(
                create,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                    "--start_date=2021-03-01",
                    f"--billing_project={VALID_BILLING_PROJECT}",
                ],
            )

            assert result.exit_code == 0
            assert BACKFILL_FILE in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            backfill_file = SQL_DIR + "/" + BACKFILL_FILE
            backfill = Backfill.entries_from_file(backfill_file)[0]

            assert backfill.entry_date == date.today()
            assert backfill.start_date == date(2021, 3, 1)
            assert backfill.end_date == date.today()
            assert backfill.watchers == [DEFAULT_WATCHER]
            assert backfill.reason == DEFAULT_REASON
            assert backfill.status == DEFAULT_STATUS
            assert backfill.billing_project == VALID_BILLING_PROJECT

    def test_create_backfill_with_invalid_billing_project_should_fail(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            result = runner.invoke(
                create,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                    "--start_date=2021-03-01",
                    f"--billing_project={INVALID_BILLING_PROJECT}",
                ],
            )

            assert result.exit_code == 1
            assert "Invalid billing project" in str(result.exception)

    def test_create_backfill_depends_on_past_should_fail(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF_DEPENDS_ON_PAST))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            result = runner.invoke(
                create,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                    "--start_date=2021-03-01",
                ],
            )

            assert result.exit_code == 1
            assert (
                "Tables that depend on past are currently not supported."
                in result.output
            )

    def test_create_backfill_with_invalid_watcher(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/test/test_query_v1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            invalid_watcher = "test.org"
            result = runner.invoke(
                create,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                    "--start_date=2021-03-01",
                    "--watcher=" + invalid_watcher,
                ],
            )
            assert result.exit_code == 1
            assert "Invalid" in str(result.exception)
            assert "watchers" in str(result.exception)

    def test_create_backfill_with_invalid_path(self, runner):
        with runner.isolated_filesystem():
            invalid_path = "test.test_query_v1"
            result = runner.invoke(create, [invalid_path, "--start_date=2021-03-01"])
            assert result.exit_code == 2
            assert "Invalid" in result.output
            assert "path" in result.output

    def test_create_backfill_with_invalid_start_date_greater_than_end_date(
        self, runner
    ):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/test/test_query_v1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            invalid_start_date = "2021-05-01"
            result = runner.invoke(
                create,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                    "--start_date=" + invalid_start_date,
                    "--end_date=2021-03-01",
                ],
            )
            assert result.exit_code == 1
            assert "Invalid start date" in str(result.exception)

    def test_create_backfill_with_invalid_excluded_dates_before_start_date(
        self, runner
    ):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/test/test_query_v1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            invalid_exclude_date = "2021-03-01"
            result = runner.invoke(
                create,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                    "--start_date=2021-05-01",
                    "--exclude=" + invalid_exclude_date,
                ],
            )
            assert result.exit_code == 1
            assert "Invalid excluded dates" in str(result.exception)

    def test_create_backfill_with_excluded_dates_after_end_date(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/test/test_query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            invalid_exclude_date = "2021-07-01"
            result = runner.invoke(
                create,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                    "--start_date=2021-05-01",
                    "--end_date=2021-06-01",
                    "--exclude=" + invalid_exclude_date,
                ],
            )
            assert result.exit_code == 1
            assert "Invalid excluded dates" in str(result.exception)

    def test_create_backfill_entry_with_all_params(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)
            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            result = runner.invoke(
                create,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                    "--start_date=2021-03-01",
                    "--end_date=2021-03-10",
                    "--exclude=2021-03-05",
                    "--watcher=test@example.org",
                ],
            )

            assert result.exit_code == 0
            assert BACKFILL_FILE in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            backfill_file = SQL_DIR + "/" + BACKFILL_FILE
            backfill = Backfill.entries_from_file(backfill_file)[0]

            assert backfill.start_date == date(2021, 3, 1)
            assert backfill.end_date == date(2021, 3, 10)
            assert backfill.watchers == [VALID_WATCHER]
            assert backfill.reason == DEFAULT_REASON
            assert backfill.status == DEFAULT_STATUS

    def test_create_backfill_with_exsting_entry(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            backfill_entry_1 = Backfill(
                date(2021, 5, 3),
                date(2021, 1, 3),
                date(2021, 5, 3),
                [date(2021, 2, 3)],
                VALID_REASON,
                [VALID_WATCHER],
                BackfillStatus.COMPLETE,
            )

            backfill_entry_2 = Backfill(
                date.today(),
                date(2023, 3, 1),
                date(2023, 3, 10),
                [],
                DEFAULT_REASON,
                [DEFAULT_WATCHER],
                DEFAULT_STATUS,
            )

            backfill_file = (
                Path("sql/moz-fx-data-shared-prod/test/test_query_v1") / BACKFILL_FILE
            )
            backfill_file.write_text(backfill_entry_1.to_yaml())
            assert BACKFILL_FILE in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            backfills = Backfill.entries_from_file(backfill_file)
            assert backfills[0] == backfill_entry_1

            result = runner.invoke(
                create,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                    "--start_date=2023-03-01",
                    "--end_date=2023-03-10",
                ],
            )

            assert result.exit_code == 0

            backfills = Backfill.entries_from_file(backfill_file)

            assert backfills[1] == backfill_entry_1
            assert backfills[0] == backfill_entry_2

    def test_create_backfill_with_existing_entry_with_initiate_status_should_fail(
        self, runner
    ):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            backfill_entry_1 = Backfill(
                date(2021, 5, 3),
                date(2021, 1, 3),
                date(2021, 5, 3),
                [date(2021, 2, 3)],
                VALID_REASON,
                [VALID_WATCHER],
                DEFAULT_STATUS,
            )

            backfill_file = (
                Path("sql/moz-fx-data-shared-prod/test/test_query_v1") / BACKFILL_FILE
            )
            backfill_file.write_text(backfill_entry_1.to_yaml())
            assert BACKFILL_FILE in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            backfills = Backfill.entries_from_file(backfill_file)
            assert backfills[0] == backfill_entry_1

            result = runner.invoke(
                create,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                    "--start_date=2023-03-01",
                    "--end_date=2023-03-10",
                ],
            )

            assert result.exit_code == 1

            assert (
                "Backfill entries cannot contain more than one entry with Initiate status"
                in str(result.exception)
            )

    def test_validate_backfill(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_file.write_text(BACKFILL_YAML_TEMPLATE)
            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 0

    def test_validate_backfill_with_billing_project(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_text = (
                BACKFILL_YAML_TEMPLATE + f"  billing_project: {VALID_BILLING_PROJECT}"
            )
            backfill_file.write_text(backfill_text)
            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 0

    def test_validate_backfill_with_invalid_billing_project_should_fail(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_text = (
                BACKFILL_YAML_TEMPLATE + f"  billing_project: {INVALID_BILLING_PROJECT}"
            )
            backfill_file.write_text(backfill_text)
            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert "Invalid billing project" in str(result.exception)

    def test_validate_backfill_depends_on_past_should_fail(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF_DEPENDS_ON_PAST))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_file.write_text(BACKFILL_YAML_TEMPLATE)
            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert (
                "Tables that depend on past are currently not supported"
                in result.output
            )

    def test_validate_backfill_invalid_table_name(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.",
                ],
            )
            assert result.exit_code == 1
            assert (
                "Qualified table name must be named like: <project>.<dataset>.<table>"
                in str(result.exception)
            )

    def test_validate_backfill_non_existing_table_name(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v2",
                ],
            )
            assert result.exit_code == 1
            assert "does not exist" in result.output

    def test_validate_backfill_invalid_default_reason(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            invalid_backfill = BACKFILL_YAML_TEMPLATE.replace(
                VALID_REASON, DEFAULT_REASON
            )
            backfill_file.write_text(invalid_backfill)
            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert "Default reason" in result.output

    def test_validate_backfill_empty_reason(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            invalid_reason = ""
            invalid_backfill = BACKFILL_YAML_TEMPLATE.replace(
                VALID_REASON, invalid_reason
            )
            backfill_file.write_text(invalid_backfill)
            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert (
                "Reason in backfill entry should not be empty"
                in result.exception.args[0]
            )

    def test_validate_backfill_invalid_watcher(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            invalid_watcher = "test@example"
            invalid_backfill = BACKFILL_YAML_TEMPLATE.replace(
                VALID_WATCHER, invalid_watcher
            )
            backfill_file.write_text(invalid_backfill)
            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert "Invalid" in str(result.exception)
            assert "watchers" in str(result.exception)

    def test_validate_backfill_empty_watcher(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            invalid_watcher = ""
            invalid_backfill = BACKFILL_YAML_TEMPLATE.replace(
                VALID_WATCHER, invalid_watcher
            )
            backfill_file.write_text(invalid_backfill)
            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert "Invalid" in str(result.exception)
            assert "watchers" in str(result.exception)

    def test_validate_backfill_watchers_duplicated(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            invalid_watchers = "  - test@example.org\n" "  - test@example.org\n"
            invalid_backfill = BACKFILL_YAML_TEMPLATE.replace(
                "  - " + VALID_WATCHER, invalid_watchers
            )
            backfill_file.write_text(invalid_backfill)
            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert "Duplicate watcher" in result.exception.args[0]

    def test_validate_backfill_invalid_status(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            invalid_status = "INVALIDSTATUS"
            invalid_backfill = BACKFILL_YAML_TEMPLATE.replace(
                DEFAULT_STATUS.value, invalid_status
            )
            backfill_file.write_text(invalid_backfill)
            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert invalid_status in str(result.exception)

    def test_validate_backfill_duplicate_entry_dates(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE

            duplicate_entry_date = "2021-05-05"
            invalid_backfill = BACKFILL_YAML_TEMPLATE.replace(
                "2021-05-04", duplicate_entry_date
            )
            backfill_file.write_text(invalid_backfill + "\n" + invalid_backfill)

            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert "Backfill entry already exists" in str(result.exception)

    def test_validate_backfill_invalid_entry_date(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            invalid_entry_date = (date.today() + timedelta(days=1)).strftime("%Y-%m-%d")
            backfill_file.write_text(
                BACKFILL_YAML_TEMPLATE.replace("2021-05-04", invalid_entry_date)
            )

            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert "can't be in the future" in str(result.exception)

    def test_validate_backfill_invalid_start_date_greater_than_end_date(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            invalid_start_date = "2021-05-04"
            backfill_file.write_text(
                BACKFILL_YAML_TEMPLATE.replace("2021-01-03", invalid_start_date)
            )

            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert "Invalid start date" in str(result.exception)

    def test_validate_backfill_invalid_start_date_greater_than_entry_date(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            invalid_start_date = "2021-05-05"
            backfill_file.write_text(
                BACKFILL_YAML_TEMPLATE.replace("2021-01-03", invalid_start_date)
            )

            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert "Invalid start date" in str(result.exception)

    def test_validate_backfill_invalid_end_date_greater_than_entry_date(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            invalid_end_date = "2021-05-05"
            backfill_file.write_text(
                BACKFILL_YAML_TEMPLATE.replace("2021-05-03", invalid_end_date)
            )

            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert "Invalid end date" in str(result.exception)

    def test_validate_backfill_invalid_excluded_dates_less_than_start_date(
        self, runner
    ):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            invalid_excluded_date = "2021-01-02"
            backfill_file.write_text(
                BACKFILL_YAML_TEMPLATE.replace("2021-02-03", invalid_excluded_date)
            )

            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert "Invalid excluded dates" in str(result.exception)

    def test_validate_backfill_invalid_excluded_dates_greater_than_end_date(
        self, runner
    ):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            invalid_excluded_date = "2021-05-04"
            backfill_file.write_text(
                BACKFILL_YAML_TEMPLATE.replace("2021-02-03", invalid_excluded_date)
            )

            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert "Invalid excluded dates" in str(result.exception)

    def test_validate_backfill_invalid_excluded_dates_duplicated(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            duplicate_excluded_dates = "  - 2021-02-03\n" "  - 2021-02-03\n"
            backfill_file.write_text(
                (
                    "2021-05-04:\n"
                    "  start_date: 2021-01-03\n"
                    "  end_date: 2021-05-03\n"
                    "  excluded_dates:\n"
                    + duplicate_excluded_dates
                    + "  reason: test_reason\n"
                    "  watchers:\n"
                    "  - test@example.org\n"
                    "  status: Initiate\n"
                )
            )

            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert "duplicate excluded dates" in result.exception.args[0]

    def test_validate_backfill_invalid_excluded_dates_not_sorted(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            duplicate_excluded_dates = "  - 2021-02-04\n" "  - 2021-02-03\n"
            backfill_file.write_text(
                "2021-05-04:\n"
                "  start_date: 2021-01-03\n"
                "  end_date: 2021-05-03\n"
                "  excluded_dates:\n"
                + duplicate_excluded_dates
                + "  reason: test_reason\n"
                "  watchers:\n"
                "  - test@example.org\n"
                "  status: Initiate\n"
            )

            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert "excluded dates not sorted" in result.exception.args[0]

    def test_validate_backfill_entries_not_sorted(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_file.write_text(
                BACKFILL_YAML_TEMPLATE + "\n"
                "2023-05-04:\n"
                "  start_date: 2020-01-03\n"
                "  end_date: 2020-05-03\n"
                "  reason: test_reason\n"
                "  watchers:\n"
                "  - test@example.org\n"
                "  status: Complete\n"
            )

            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                validate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )
            assert result.exit_code == 1
            assert "entries are not sorted" in result.output

    def test_backfill_info_one_table_all_status(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            qualified_table_name_1 = "moz-fx-data-shared-prod.test.test_query_v1"

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_file.write_text(
                BACKFILL_YAML_TEMPLATE + "\n"
                "2021-05-03:\n"
                "  start_date: 2021-01-03\n"
                "  end_date: 2021-05-03\n"
                "  reason: test_reason\n"
                "  watchers:\n"
                "  - test@example.org\n"
                "  status: Complete\n"
            )

            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v2"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v2/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v2/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            qualified_table_name_2 = "moz-fx-data-shared-prod.test.test_query_v2"

            result = runner.invoke(
                create,
                [
                    qualified_table_name_2,
                    "--start_date=2021-04-01",
                ],
            )

            assert result.exit_code == 0
            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                info,
                [qualified_table_name_1],
            )

            assert result.exit_code == 0
            assert qualified_table_name_1 in result.output
            assert BackfillStatus.INITIATE.value in result.output
            assert "total of 2 backfill(s)" in result.output
            assert qualified_table_name_2 not in result.output
            assert BackfillStatus.COMPLETE.value in result.output

    def test_backfill_info_one_table_initiate_status(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)
            qualified_table_name = "moz-fx-data-shared-prod.test.test_query_v1"

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_file.write_text(
                BACKFILL_YAML_TEMPLATE + "\n"
                "2021-05-03:\n"
                "  start_date: 2021-01-03\n"
                "  end_date: 2021-05-03\n"
                "  reason: test_reason\n"
                "  watchers:\n"
                "  - test@example.org\n"
                "  status: Initiate\n"
            )

            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                info,
                [qualified_table_name, "--status=Initiate"],
            )

            assert result.exit_code == 0
            assert qualified_table_name in result.output
            assert BackfillStatus.INITIATE.value in result.output
            assert "total of 2 backfill(s)" in result.output
            assert BackfillStatus.COMPLETE.value not in result.output

    def test_backfill_info_all_tables_all_status(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            qualified_table_name_1 = "moz-fx-data-shared-prod.test.test_query_v1"

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_file.write_text(
                BACKFILL_YAML_TEMPLATE + "\n"
                "2021-05-03:\n"
                "  start_date: 2021-01-03\n"
                "  end_date: 2021-05-03\n"
                "  reason: test_reason\n"
                "  watchers:\n"
                "  - test@example.org\n"
                "  status: Complete\n"
            )

            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v2"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v2/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v2/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            qualified_table_name_2 = "moz-fx-data-shared-prod.test.test_query_v2"

            result = runner.invoke(
                create,
                [
                    qualified_table_name_2,
                    "--start_date=2021-04-01",
                ],
            )

            assert result.exit_code == 0
            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                info,
                [],
            )

            assert result.exit_code == 0
            assert qualified_table_name_1 in result.output
            assert qualified_table_name_2 in result.output
            assert BackfillStatus.INITIATE.value in result.output
            assert "total of 3 backfill(s)" in result.output
            assert BackfillStatus.COMPLETE.value in result.output

    def test_backfill_info_all_tables_with_initiate_status(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            qualified_table_name_1 = "moz-fx-data-shared-prod.test.test_query_v1"
            os.makedirs(SQL_DIR)

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_file.write_text(
                BACKFILL_YAML_TEMPLATE + "\n"
                "2021-05-03:\n"
                "  start_date: 2021-01-03\n"
                "  end_date: 2021-05-03\n"
                "  reason: test_reason\n"
                "  watchers:\n"
                "  - test@example.org\n"
                "  status: Initiate\n"
            )

            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v2"
            qualified_table_name_2 = "moz-fx-data-shared-prod.test.test_query_v2"
            os.makedirs(SQL_DIR)

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_file.write_text(
                BACKFILL_YAML_TEMPLATE + "\n"
                "2021-05-03:\n"
                "  start_date: 2021-01-03\n"
                "  end_date: 2021-05-03\n"
                "  reason: test_reason\n"
                "  watchers:\n"
                "  - test@example.org\n"
                "  status: Initiate\n"
            )

            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                info,
                [
                    "--status=Initiate",
                ],
            )

            assert result.exit_code == 0
            assert qualified_table_name_1 in result.output
            assert qualified_table_name_2 in result.output
            assert BackfillStatus.INITIATE.value in result.output
            assert "total of 4 backfill(s)" in result.output
            assert BackfillStatus.COMPLETE.value not in result.output

    def test_backfill_info_with_invalid_path(self, runner):
        with runner.isolated_filesystem():
            invalid_path = "moz-fx-data-shared-prod.test.test_query_v2"
            result = runner.invoke(info, [invalid_path])

            assert result.exit_code == 2
            assert "Invalid" in result.output
            assert "path" in result.output

    def test_backfill_info_with_invalid_qualified_table_name(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)
            invalid_qualified_table_name = "mozdata.test_query_v2"
            result = runner.invoke(info, [invalid_qualified_table_name])

            assert result.exit_code == 1
            assert "Qualified table name must be named like:" in str(result.exception)

    def test_backfill_info_one_table_invalid_status(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)
            qualified_table_name = "moz-fx-data-shared-prod.test.test_query_v1"

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_file.write_text(
                BACKFILL_YAML_TEMPLATE + "\n"
                "2021-05-03:\n"
                "  start_date: 2021-01-03\n"
                "  end_date: 2021-05-03\n"
                "  reason: test_reason\n"
                "  watchers:\n"
                "  - test@example.org\n"
                "  status: Complete\n"
            )

            assert BACKFILL_FILE in os.listdir(SQL_DIR)

            result = runner.invoke(
                info,
                [qualified_table_name, "--status=testing"],
            )

            assert result.exit_code == 2
            assert "Invalid value for '--status'" in result.output

    def test_get_entries_from_qualified_table_name(self, runner):
        with runner.isolated_filesystem():
            qualified_table_name = "moz-fx-data-shared-prod.test.test_query_v1"

            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            result = runner.invoke(
                create,
                [
                    qualified_table_name,
                    "--start_date=2021-03-01",
                ],
            )

            assert result.exit_code == 0
            assert BACKFILL_FILE in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            backfill_file = SQL_DIR + "/" + BACKFILL_FILE
            backfill = Backfill.entries_from_file(backfill_file)[0]

            assert backfill.entry_date == date.today()
            assert backfill.start_date == date(2021, 3, 1)
            assert backfill.end_date == date.today()
            assert backfill.watchers == [DEFAULT_WATCHER]
            assert backfill.reason == DEFAULT_REASON
            assert backfill.status == DEFAULT_STATUS

            backfills = get_entries_from_qualified_table_name(
                "sql", qualified_table_name
            )

            expected_backfill = Backfill(
                date.today(),
                date(2021, 3, 1),
                date.today(),
                [],
                DEFAULT_REASON,
                [DEFAULT_WATCHER],
                DEFAULT_STATUS,
            )

            assert backfills == [expected_backfill]

    def test_get_qualified_table_name_to_entries_map_by_project(self, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            qualified_table_name_1 = "moz-fx-data-shared-prod.test.test_query_v1"

            result = runner.invoke(
                create,
                [
                    qualified_table_name_1,
                    "--start_date=2021-03-01",
                ],
            )

            assert result.exit_code == 0
            assert BACKFILL_FILE in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            qualified_table_name_2 = "moz-fx-data-shared-prod.test.test_query_v2"
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v2"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v2/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v2/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            result = runner.invoke(
                create,
                [
                    qualified_table_name_2,
                    "--start_date=2021-03-01",
                ],
            )

            assert result.exit_code == 0
            assert BACKFILL_FILE in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v2"
            )

            backfills_dict = get_qualified_table_name_to_entries_map_by_project(
                "sql", "moz-fx-data-shared-prod"
            )

            expected_backfill = Backfill(
                date.today(),
                date(2021, 3, 1),
                date.today(),
                [],
                DEFAULT_REASON,
                [DEFAULT_WATCHER],
                DEFAULT_STATUS,
            )

            assert qualified_table_name_1 in backfills_dict
            assert backfills_dict[qualified_table_name_1] == [expected_backfill]
            assert qualified_table_name_2 in backfills_dict

    def test_get_backfill_file_from_qualified_table_name(self, runner):
        qualified_table_name = "moz-fx-data-shared-prod.test.test_query_v1"
        actual_backfill_file = get_backfill_file_from_qualified_table_name(
            "sql", qualified_table_name
        )
        expected_backfill_file = Path(
            "sql/moz-fx-data-shared-prod/test/test_query_v1/backfill.yaml"
        )

        assert actual_backfill_file == expected_backfill_file

    def test_get_backfill_staging_qualified_table_name(self, runner):
        qualified_table_name = "moz-fx-data-shared-prod.test.test_query_v1"
        backfill_table_id = "test__test_query_v1_2023_05_30"

        actual_backfill_staging = get_backfill_staging_qualified_table_name(
            qualified_table_name, date.fromisoformat("2023-05-30")
        )
        expected_backfill_staging = f"{BACKFILL_DESTINATION_PROJECT}.{BACKFILL_DESTINATION_DATASET}.{backfill_table_id}"

        assert actual_backfill_staging == expected_backfill_staging

    def test_get_backfill_backup_table_name(self, runner):
        qualified_table_name = "moz-fx-data-shared-prod.test.test_query_v1"
        cloned_table_id = "test__test_query_v1_backup_2023_05_30"

        actual_backfill_staging = get_backfill_backup_table_name(
            qualified_table_name, date.fromisoformat("2023-05-30")
        )
        expected_backfill_backup = f"{BACKFILL_DESTINATION_PROJECT}.{BACKFILL_DESTINATION_DATASET}.{cloned_table_id}"

        assert actual_backfill_staging == expected_backfill_backup

    def test_validate_metadata_workgroups_invalid_table_workgroup_and_valid_dataset_workgroup(
        self, runner
    ):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/test/test_query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            assert "query.sql" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            qualified_table_name = "moz-fx-data-shared-prod.test.test_query_v1"

            invalid_workgroup_access = [
                dict(
                    role="roles/bigquery.dataViewer",
                    members=["workgroup:invalid_workgroup"],
                )
            ]

            metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "owners": ["test@example.org"],
                "workgroup_access": invalid_workgroup_access,
            }

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))

            assert "metadata.yaml" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            result = validate_metadata_workgroups("sql", qualified_table_name)
            assert result

    def test_validate_metadata_workgroups_empty_table_workgroup_and_valid_dataset_workgroups(
        self, runner
    ):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/test/test_query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            assert "query.sql" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            qualified_table_name = "moz-fx-data-shared-prod.test.test_query_v1"

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF_EMPTY_WORKGROUP))

            assert "metadata.yaml" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            result = validate_metadata_workgroups("sql", qualified_table_name)
            assert result

    def test_validate_metadata_workgroups_valid_table_workgroup_and_invalid_dataset_workgroup(
        self, runner
    ):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/test/test_query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            assert "query.sql" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            qualified_table_name = "moz-fx-data-shared-prod.test.test_query_v1"

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            assert "metadata.yaml" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            invalid_workgroup_access = [
                dict(
                    role="roles/bigquery.dataViewer",
                    members=["workgroup:invalid_workgroup"],
                )
            ]

            dataset_metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "dataset_base_acl": "derived",
                "workgroup_access": invalid_workgroup_access,
            }

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(dataset_metadata_conf))

            assert "dataset_metadata.yaml" in os.listdir(
                "sql/moz-fx-data-shared-prod/test"
            )

            result = validate_metadata_workgroups("sql", qualified_table_name)
            assert result

    def test_validate_metadata_workgroups_valid_table_workgroup_and_empty_dataset_workgroup(
        self, runner
    ):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/test/test_query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            assert "query.sql" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            qualified_table_name = "moz-fx-data-shared-prod.test.test_query_v1"

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            assert "metadata.yaml" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF_EMPTY_WORKGROUP))

            assert "dataset_metadata.yaml" in os.listdir(
                "sql/moz-fx-data-shared-prod/test"
            )

            result = validate_metadata_workgroups("sql", qualified_table_name)
            assert result

    def test_validate_metadata_workgroups_missing_table_metadata(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/test/test_query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            assert "query.sql" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            qualified_table_name = "moz-fx-data-shared-prod.test.test_query_v1"

            assert "metadata.yaml" not in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF_EMPTY_WORKGROUP))

            assert "dataset_metadata.yaml" in os.listdir(
                "sql/moz-fx-data-shared-prod/test"
            )

            result = validate_metadata_workgroups("sql", qualified_table_name)
            assert result

    def test_validate_metadata_workgroups_missing_dataset_metadata(self, runner):
        with pytest.raises(ValueError) as e:
            with runner.isolated_filesystem():
                os.makedirs("sql/moz-fx-data-shared-prod/test/test_query_v1")
                with open(
                    "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
                ) as f:
                    f.write("SELECT 1")

                assert "query.sql" in os.listdir(
                    "sql/moz-fx-data-shared-prod/test/test_query_v1"
                )

                qualified_table_name = "moz-fx-data-shared-prod.test.test_query_v1"

                with open(
                    "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                    "w",
                ) as f:
                    f.write(yaml.dump(TABLE_METADATA_CONF))

                assert "metadata.yaml" in os.listdir(
                    "sql/moz-fx-data-shared-prod/test/test_query_v1"
                )

                assert "dataset_metadata.yaml" not in os.listdir(
                    "sql/moz-fx-data-shared-prod/test"
                )

                validate_metadata_workgroups("sql", qualified_table_name)

        assert e.type == ValueError

    def test_validate_metadata_workgroups_empty_workgroups(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/test/test_query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            assert "query.sql" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            qualified_table_name = "moz-fx-data-shared-prod.test.test_query_v1"

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF_EMPTY_WORKGROUP))

            assert "metadata.yaml" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF_EMPTY_WORKGROUP))

            assert "dataset_metadata.yaml" in os.listdir(
                "sql/moz-fx-data-shared-prod/test"
            )

            result = validate_metadata_workgroups("sql", qualified_table_name)
            assert not result

    def test_qualified_table_name_matching(self, runner):
        qualified_table_name = "moz-fx-data-shared-prod.test.test_query_v1"
        project_id, dataset_id, table_id = qualified_table_name_matching(
            qualified_table_name
        )

        assert project_id == "moz-fx-data-shared-prod"
        assert dataset_id == "test"
        assert table_id == "test_query_v1"

    def test_qualified_table_name_matching_invalid_name(self, runner):
        with pytest.raises(AttributeError) as e:
            invalid_name = "test.test_query_v1"
            qualified_table_name_matching(invalid_name)

        assert "Qualified table name must be named like" in str(e.value)

    @patch("google.cloud.bigquery.Client.get_table")
    def test_backfill_scheduled(self, get_table, runner):
        get_table.side_effect = [
            None,  # Check that staging data exists
            NotFound(  # Check that clone does not exist
                "moz-fx-data-shared-prod.backfills_staging_derived.test_query_v1_backup_2021_05_03"
                "not found"
            ),
            NotFound(  # Check that staging data does not exist
                "moz-fx-data-shared-prod.backfills_staging_derived.test_query_v1_2021_05_04"
                "not found"
            ),
        ]
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_file.write_text(
                BACKFILL_YAML_TEMPLATE
                + """
2021-05-03:
  start_date: 2021-01-03
  end_date: 2021-05-03
  reason: test_reason
  watchers:
  - test@example.org
  status: Complete"""
            )

            result = runner.invoke(
                scheduled,
                [
                    "--json_path=tmp.json",
                    "--status=Complete",
                ],
            )

            assert result.exit_code == 0
            assert "1 backfill(s) require processing." in result.output
            assert Path("tmp.json").exists()
            assert len(json.loads(Path("tmp.json").read_text())) == 1

            result = runner.invoke(
                scheduled,
                [
                    "--json_path=tmp.json",
                    "--status=Initiate",
                ],
            )

            assert result.exit_code == 0
            assert "1 backfill(s) require processing." in result.output

    @patch("google.cloud.bigquery.Client.get_table")
    def test_backfill_scheduled_depends_on_past_should_fail(self, get_table, runner):
        get_table.side_effect = [
            None,  # Check that staging data exists
            NotFound(  # Check that clone does not exist
                "moz-fx-data-shared-prod.backfills_staging_derived.test_query_v1_backup_2021_05_03"
                "not found"
            ),
            NotFound(  # Check that staging data does not exist
                "moz-fx-data-shared-prod.backfills_staging_derived.test_query_v1_2021_05_04"
                "not found"
            ),
        ]
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF_DEPENDS_ON_PAST))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_file.write_text(
                BACKFILL_YAML_TEMPLATE
                + """
2021-05-03:
  start_date: 2021-01-03
  end_date: 2021-05-03
  reason: test_reason
  watchers:
  - test@example.org
  status: Complete"""
            )

            result = runner.invoke(
                scheduled,
                [
                    "--json_path=tmp.json",
                    "--status=Complete",
                ],
            )

            assert result.exit_code == 0
            assert "0 backfill(s) require processing." in result.output
            assert Path("tmp.json").exists()
            assert len(json.loads(Path("tmp.json").read_text())) == 0

            result = runner.invoke(
                scheduled,
                [
                    "--json_path=tmp.json",
                    "--status=Initiate",
                ],
            )

            assert result.exit_code == 0
            assert "0 backfill(s) require processing." in result.output

    @patch("google.cloud.bigquery.Client.get_table")
    @patch("google.cloud.bigquery.Client.copy_table")
    @patch("google.cloud.bigquery.Client.delete_table")
    def test_complete_partitioned_backfill(
        self, delete_table, copy_table, get_table, runner
    ):
        get_table.side_effect = [
            None,  # Check that staging data exists
            NotFound(  # Check that clone does not exist
                "moz-fx-data-shared-prod.backfills_staging_derived.test__test_query_v1_backup_2021_05_03"
                "not found"
            ),
        ]
        copy_table.side_effect = None
        delete_table.side_effect = None

        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(PARTITIONED_TABLE_METADATA))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF_EMPTY_WORKGROUP))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_file.write_text(
                """
2021-05-03:
  start_date: 2021-01-03
  end_date: 2021-01-13
  reason: test_reason
  watchers:
  - test@example.org
  status: Complete"""
            )

            result = runner.invoke(
                complete,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                ],
            )

            assert result.exit_code == 0
            assert copy_table.call_count == 12  # one for backup, 11 for partitions
            for i, call in enumerate(copy_table.call_args_list[1:]):
                d = date(2021, 1, 3) + timedelta(days=i)
                assert call.args == (
                    f'moz-fx-data-shared-prod.backfills_staging_derived.test__test_query_v1_2021_05_03${d.strftime("%Y%m%d")}',
                    f'moz-fx-data-shared-prod.test.test_query_v1${d.strftime("%Y%m%d")}',
                )
            assert delete_table.call_count == 1

    @patch("google.cloud.bigquery.Client")
    @patch("subprocess.check_call")
    @patch("bigquery_etl.cli.backfill.deploy_table")
    @patch("bigquery_etl.cli.backfill.Schema.from_query_file")
    def test_initiate_partitioned_backfill(
        self,
        mock_from_query_file,
        mock_deploy_table,
        check_call,
        mock_client,
        runner,
    ):
        backfill_staging_table_name = (
            "moz-fx-data-shared-prod.backfills_staging_derived.test__test_query_v1"
        )

        mock_client().get_table.side_effect = [
            NotFound(  # Check that staging data does not exist
                f"{backfill_staging_table_name}_backup_2021_05_03" "not found"
            ),
            None,  # Check that production data exists during dry run
            None,  # Check that production data exists
        ]

        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)
            query_path = Path(SQL_DIR) / QUERY_FILE

            with open(query_path, "w") as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/schema.yaml", "w"
            ) as f:
                f.write(yaml.dump({"fields": [{"name": "f0_", "type": "INTEGER"}]}))

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(PARTITIONED_TABLE_METADATA))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF_EMPTY_WORKGROUP))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_file.write_text(
                """
            2021-05-03:
              start_date: 2021-01-03
              end_date: 2021-01-08
              reason: test_reason
              watchers:
              - test@example.org
              status: Initiate
              override_retention_limit: True"""
            )

            result = runner.invoke(
                initiate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                    "--parallelism=0",
                ],
            )

            assert result.exit_code == 0

            assert not mock_from_query_file.called  # schema exists

            mock_deploy_table.assert_called_with(
                artifact_file=query_path,
                destination_table=f"{backfill_staging_table_name}_2021_05_03",
                respect_dryrun_skip=False,
            )

            expected_submission_date_params = [
                f"--parameter=submission_date:DATE:2021-01-0{day}"
                for day in range(3, 9)
            ]

            expected_destination_table_params = [
                f"--destination_table=moz-fx-data-shared-prod:backfills_staging_derived.test__test_query_v1_2021_05_03$2021010{day}"
                for day in range(3, 9)
            ]

            # this is inspecting calls to the underlying subprocess.check_call(["bq]"...)
            assert check_call.call_count == 12  # 6 for dry run, 6 for backfill
            for call in check_call.call_args_list:
                submission_date_params = [
                    arg for arg in call.args[0] if "--parameter=submission_date" in arg
                ]
                assert len(submission_date_params) == 1
                assert submission_date_params[0] in expected_submission_date_params
                assert f"--project_id={DEFAULT_BILLING_PROJECT}" in call.args[0]
                destination_table_params = [
                    arg for arg in call.args[0] if "--destination_table" in arg
                ]
                assert destination_table_params[0] in expected_destination_table_params

    @patch("google.cloud.bigquery.Client")
    @patch("subprocess.check_call")
    @patch("bigquery_etl.cli.backfill.deploy_table")
    @patch("bigquery_etl.cli.backfill.Schema.from_query_file")
    def test_initiate_partitioned_backfill_without_schema_should_pass(
        self, mock_from_query_file, mock_deploy_table, check_call, mock_client, runner
    ):
        backfill_staging_table_name = (
            "moz-fx-data-shared-prod.backfills_staging_derived.test__test_query_v1"
        )

        mock_client().get_table.side_effect = [
            NotFound(  # Check that staging data does not exist
                f"{backfill_staging_table_name}_backup_2021_05_03" "not found"
            ),
            None,  # Check that production data exists during dry run
            None,  # Check that production data exists
        ]

        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)
            query_path = Path(SQL_DIR) / QUERY_FILE

            with open(query_path, "w") as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(PARTITIONED_TABLE_METADATA))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF_EMPTY_WORKGROUP))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_file.write_text(
                """
            2021-05-03:
              start_date: 2021-01-03
              end_date: 2021-01-08
              reason: test_reason
              watchers:
              - test@example.org
              status: Initiate
              override_retention_limit: true"""
            )

            result = runner.invoke(
                initiate,
                ["moz-fx-data-shared-prod.test.test_query_v1", "--parallelism=0"],
            )

            assert result.exit_code == 0

            mock_from_query_file.assert_called_with(
                query_file=query_path,
                respect_skip=False,
                sql_dir="sql/",
            )

            mock_deploy_table.assert_called_with(
                artifact_file=query_path,
                destination_table=f"{backfill_staging_table_name}_2021_05_03",
                respect_dryrun_skip=False,
            )

            expected_submission_date_params = [
                f"--parameter=submission_date:DATE:2021-01-0{day}"
                for day in range(3, 9)
            ]

            expected_destination_table_params = [
                f"--destination_table=moz-fx-data-shared-prod:backfills_staging_derived.test__test_query_v1_2021_05_03$2021010{day}"
                for day in range(3, 9)
            ]

            # this is inspecting calls to the underlying subprocess.check_call(["bq]"...)
            assert check_call.call_count == 12  # 6 for dry run, 6 for backfill
            for call in check_call.call_args_list:
                submission_date_params = [
                    arg for arg in call.args[0] if "--parameter=submission_date" in arg
                ]
                assert len(submission_date_params) == 1
                assert submission_date_params[0] in expected_submission_date_params
                assert f"--project_id={DEFAULT_BILLING_PROJECT}" in call.args[0]
                destination_table_params = [
                    arg for arg in call.args[0] if "--destination_table" in arg
                ]
                assert destination_table_params[0] in expected_destination_table_params

    @patch("google.cloud.bigquery.Client")
    @patch("subprocess.check_call")
    @patch("bigquery_etl.cli.backfill.deploy_table")
    @patch("bigquery_etl.cli.backfill.Schema.from_query_file")
    def test_initiate_partitioned_backfill_with_valid_billing_project_from_entry(
        self, mock_from_query_file, mock_deploy_table, check_call, mock_client, runner
    ):
        backfill_staging_table_name = (
            "moz-fx-data-shared-prod.backfills_staging_derived.test__test_query_v1"
        )

        mock_client().get_table.side_effect = [
            NotFound(  # Check that staging data does not exist
                f"{backfill_staging_table_name}_backup_2021_05_03" "not found"
            ),
            None,  # Check that production data exists during dry run
            None,  # Check that production data exists
        ]

        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)
            query_path = Path(SQL_DIR) / QUERY_FILE

            with open(query_path, "w") as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/schema.yaml", "w"
            ) as f:
                f.write(yaml.dump({"fields": [{"name": "f0_", "type": "INTEGER"}]}))

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(PARTITIONED_TABLE_METADATA))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF_EMPTY_WORKGROUP))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_file.write_text(
                f"""
            2021-05-03:
              start_date: 2021-01-03
              end_date: 2021-01-08
              reason: test_reason
              watchers:
              - test@example.org
              status: Initiate
              billing_project: {VALID_BILLING_PROJECT}
              override_retention_limit: true
              """
            )

            result = runner.invoke(
                initiate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                    "--parallelism=0",
                ],
            )

            assert not mock_from_query_file.called  # schema exists

            mock_deploy_table.assert_called_with(
                artifact_file=query_path,
                destination_table=f"{backfill_staging_table_name}_2021_05_03",
                respect_dryrun_skip=False,
            )
            assert result.exit_code == 0

            expected_submission_date_params = [
                f"--parameter=submission_date:DATE:2021-01-0{day}"
                for day in range(3, 9)
            ]

            expected_destination_table_params = [
                f"--destination_table=moz-fx-data-shared-prod:backfills_staging_derived.test__test_query_v1_2021_05_03$2021010{day}"
                for day in range(3, 9)
            ]

            # this is inspecting calls to the underlying subprocess.check_call(["bq]"...)
            assert check_call.call_count == 12  # 6 for dry run, 6 for backfill
            for call in check_call.call_args_list:
                submission_date_params = [
                    arg for arg in call.args[0] if "--parameter=submission_date" in arg
                ]
                assert len(submission_date_params) == 1
                assert submission_date_params[0] in expected_submission_date_params
                assert f"--project_id={VALID_BILLING_PROJECT}" in call.args[0]
                destination_table_params = [
                    arg for arg in call.args[0] if "--destination_table" in arg
                ]
                assert destination_table_params[0] in expected_destination_table_params

    @patch("google.cloud.bigquery.Client")
    @patch("bigquery_etl.cli.backfill.deploy_table")
    @patch("bigquery_etl.cli.backfill.Schema.from_query_file")
    def test_initiate_backfill_with_failed_deploy(
        self, mock_from_query_file, mock_deploy_table, mock_client, runner
    ):
        backfill_staging_table_name = (
            "moz-fx-data-shared-prod.backfills_staging_derived.test__test_query_v1"
        )

        mock_deploy_table.side_effect = FailedDeployException("Unable to deploy table")

        mock_client().get_table.side_effect = [
            NotFound(  # Check that staging data does not exist
                f"{backfill_staging_table_name}_backup_2021_05_03" "not found"
            ),
            None,  # Check that production data exists during dry run
            None,  # Check that production data exists
        ]

        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)
            query_path = Path(SQL_DIR) / QUERY_FILE

            with open(query_path, "w") as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(PARTITIONED_TABLE_METADATA))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF_EMPTY_WORKGROUP))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_file.write_text(
                """
            2021-05-03:
              start_date: 2021-01-03
              end_date: 2021-01-08
              reason: test_reason
              watchers:
              - test@example.org
              status: Initiate
              """
            )

            result = runner.invoke(
                initiate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                    "--parallelism=0",
                ],
            )

            mock_from_query_file.assert_called_with(
                query_file=query_path,
                respect_skip=False,
                sql_dir="sql/",
            )

            mock_deploy_table.assert_called_with(
                artifact_file=query_path,
                destination_table=f"{backfill_staging_table_name}_2021_05_03",
                respect_dryrun_skip=False,
            )
            assert result.exit_code == 1
            assert "Backfill initiate failed to deploy" in str(result.exception)

    @patch("google.cloud.bigquery.Client")
    def test_initiate_partitioned_backfill_with_invalid_billing_project_from_entry_should_fail(
        self, mock_client, runner
    ):
        backfill_staging_table_name = (
            "moz-fx-data-shared-prod.backfills_staging_derived.test__test_query_v1"
        )

        mock_client().get_table.side_effect = [
            NotFound(  # Check that staging data does not exist
                f"{backfill_staging_table_name}_backup_2021_05_03" "not found"
            ),
            None,  # Check that production data exists during dry run
            None,  # Check that production data exists
        ]

        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_query_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(PARTITIONED_TABLE_METADATA))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF_EMPTY_WORKGROUP))

            backfill_file = Path(SQL_DIR) / BACKFILL_FILE
            backfill_file.write_text(
                f"""
            2021-05-03:
              start_date: 2021-01-03
              end_date: 2021-01-08
              reason: test_reason
              watchers:
              - test@example.org
              status: Initiate
              billing_project: {INVALID_BILLING_PROJECT}
              """
            )

            result = runner.invoke(
                initiate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                    "--parallelism=0",
                ],
            )

            assert result.exit_code == 1
            assert "Invalid billing project" in str(result.exception)

    @patch("bigquery_etl.cli.backfill.deploy_table")
    @patch("bigquery_etl.backfill.utils._should_initiate")
    def test_dont_initiate_if_label_and_backfill_entry_dont_match(
        self, mock_should_initiate, mock_deploy_table, runner
    ):
        """Test that process stops if the table has label shredder_mitigation and the backfill doesn't."""
        path = Path("sql/moz-fx-data-shared-prod/test/test_query_v1")
        mock_should_initiate.return_value = True
        mock_deploy_table.return_value = None
        expected_error_output = (
            "This backfill cannot continue.\nManaged backfills for "
            "tables with metadata label shredder_mitigation require "
            "using --shredder_mitigation."
        )

        with runner.isolated_filesystem():
            os.makedirs(path, exist_ok=True)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(path / "metadata.yaml", "w") as f:
                f.write(
                    "friendly_name: Test\ndescription: Test\nlabels:\n  incremental: true\n  shredder_mitigation: true"
                )

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            backfill_entry_1 = Backfill(
                date(2021, 5, 3),
                date(2021, 1, 3),
                date(2021, 5, 3),
                [date(2021, 2, 3)],
                VALID_REASON,
                [VALID_WATCHER],
                BackfillStatus.INITIATE,
                shredder_mitigation=False,
            )

            backfill_file = (
                Path("sql/moz-fx-data-shared-prod/test/test_query_v1") / BACKFILL_FILE
            )
            backfill_file.write_text(backfill_entry_1.to_yaml())
            assert BACKFILL_FILE in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )
            backfills = Backfill.entries_from_file(backfill_file)
            assert backfills[0] == backfill_entry_1

            result = runner.invoke(
                initiate,
                [
                    "moz-fx-data-shared-prod.test.test_query_v1",
                    "--parallelism=0",
                ],
            )

            assert result.exit_code == 1
            assert expected_error_output in result.output

    @patch("bigquery_etl.cli.backfill.deploy_table")
    @patch("bigquery_etl.backfill.utils._should_initiate")
    def test_initiate_if_label_and_backfill_entry_match(
        self, mock_should_initiate, mock_deploy_table, runner
    ):
        """Test that the process continues if both table & backfill use shredder_mitigation."""
        path = Path("sql/moz-fx-data-shared-prod/test/test_query_v1")
        mock_should_initiate.return_value = True
        mock_deploy_table.return_value = None

        with runner.isolated_filesystem():
            os.makedirs(path, exist_ok=True)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(path / "metadata.yaml", "w") as f:
                f.write(
                    "friendly_name: Test\ndescription: Test\nlabels:\n  incremental: true\n  shredder_mitigation: true"
                )

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            backfill_entry_1 = Backfill(
                date(2021, 5, 3),
                date(2021, 1, 3),
                date(2021, 5, 3),
                [date(2021, 2, 3)],
                VALID_REASON,
                [VALID_WATCHER],
                BackfillStatus.INITIATE,
                shredder_mitigation=True,
            )

            backfill_file = (
                Path("sql/moz-fx-data-shared-prod/test/test_query_v1") / BACKFILL_FILE
            )
            backfill_file.write_text(backfill_entry_1.to_yaml())
            assert BACKFILL_FILE in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )
            backfills = Backfill.entries_from_file(backfill_file)
            assert backfills[0] == backfill_entry_1

            with patch(
                "bigquery_etl.cli.backfill.query_backfill", return_value=None
            ) as mock_backfill:
                with patch(
                    "bigquery_etl.cli.backfill.generate_query_with_shredder_mitigation",
                    return_value=("a", "b"),
                ) as mock_shredder_mitigation:
                    result = runner.invoke(
                        initiate,
                        [
                            "moz-fx-data-shared-prod.test.test_query_v1",
                            "--parallelism=0",
                        ],
                    )

                assert result.exit_code == 0
                mock_shredder_mitigation.call_count == 2
                mock_backfill.call_count == 2



import subprocess
import sys
import textwrap

import pytest


def test_basic():
    results = subprocess.check_output(
        [
            sys.executable,
            "sql/moz-fx-data-shared-prod/telemetry/longitudinal.sql.py",
            "--tablename",
            "foo",
            "--to",
            "20190401",
        ]
    ).decode("ascii")
    assert results == textwrap.dedent(
        """
      WITH aggregated AS (
        SELECT
         ROW_NUMBER() OVER (PARTITION BY client_id ORDER BY submission_date_s3) AS _n,
         client_id,
         ARRAY_AGG(STRUCT(*)) OVER (PARTITION BY client_id ORDER BY submission_date_s3 ROWS BETWEEN CURRENT ROW AND UNBOUNDED FOLLOWING) AS aggcol
        FROM
          foo
        WHERE
          submission_date_s3 > '2018-10-01'
          AND submission_date_s3 <= '2019-04-01')
      SELECT
        * except (_n)
      FROM
        aggregated
      WHERE
        _n = 1

        """  # noqa: E501
    )


def test_everything():
    results = subprocess.check_output(
        [
            sys.executable,
            "sql/moz-fx-data-shared-prod/telemetry/longitudinal.sql.py",
            "--tablename",
            "foo",
            "--to",
            "20190401",
            "--submission-date-col",
            "submission_date_baz",
            "--select",
            "* EXCEPT locale, active_addons",
            "--where",
            "normalized_channel = 'beta'",
            "--grouping-column",
            "os",
            "--ordering-columns",
            "submission_date,app_name",
            "--max-array-length",
            "10000",
        ]
    ).decode("ascii")
    assert results == textwrap.dedent(
        """
      WITH aggregated AS (
        SELECT
         ROW_NUMBER() OVER (PARTITION BY os ORDER BY submission_date, app_name) AS _n,
         os,
         ARRAY_AGG(STRUCT(* EXCEPT locale, active_addons)) OVER (PARTITION BY os ORDER BY submission_date, app_name ROWS BETWEEN CURRENT ROW AND 10000 FOLLOWING) AS aggcol
        FROM
          foo
        WHERE
          submission_date_baz > '2018-10-01'
          AND submission_date_baz <= '2019-04-01'
          AND normalized_channel = 'beta')
      SELECT
        * except (_n)
      FROM
        aggregated
      WHERE
        _n = 1

        """  # noqa: E501
    )


def test_column_conflict():
    with pytest.raises(subprocess.CalledProcessError) as e:
        subprocess.check_output(
            [
                sys.executable,
                "sql/moz-fx-data-shared-prod/telemetry/longitudinal.sql.py",
                "--tablename",
                "foo",
                "--to",
                "20190401",
                "--grouping-column",
                "os",
                "--ordering-columns",
                "os,app_name",
            ],
            stderr=subprocess.STDOUT,
        )
    assert e.value.output.decode("ascii").endswith(
        "'os' can't be used as both a grouping and ordering column\n"
    )



"""Extract query results and write the combined JSON to a single file."""

import json
import logging
from datetime import datetime, timedelta, timezone

import rich_click as click
from google.cloud import storage  # type: ignore
from google.cloud import bigquery

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


@click.command()
@click.option(
    "--source-project",
    required=True,
    help="Google Cloud Project where the source table is located.",
)
@click.option(
    "--source-dataset",
    required=True,
    help="Dataset in BigQuery where the source table is located.",
)
@click.option(
    "--source-table", required=True, help="Name of the source table in BigQuery."
)
@click.option(
    "--destination-bucket",
    required=True,
    help="Destination Google Cloud Storage Bucket.",
)
@click.option(
    "--destination-prefix", required=True, help="Prefix of the bucket path in GCS."
)
@click.option(
    "--destination-prefix", required=True, help="Prefix of the bucket path in GCS."
)
@click.option(
    "--deletion-days-old",
    required=True,
    type=int,
    help="Number of days after which files in GCS should be deleted.",
)
def export_newtab_merino_table_to_gcs(
    source_project: str,
    source_dataset: str,
    source_table: str,
    destination_bucket: str,
    destination_prefix: str,
    deletion_days_old: int,
):
    """Use bigquery client to export data from BigQuery to GCS."""
    client = bigquery.Client(source_project)
    error_counter = 0
    threshold = 1

    try:
        # Generate the current timestamp
        timestamp = datetime.utcnow().strftime("%Y%m%d%H%M")

        # BigQuery does not export the proper JSON format, so we use a temp file and reformat
        temp_file = "temp.ndjson"

        job_config = bigquery.job.ExtractJobConfig(
            destination_format=bigquery.job.DestinationFormat.NEWLINE_DELIMITED_JSON
        )

        destination_uri = f"gs://{destination_bucket}/{destination_prefix}/{temp_file}"

        extract_job = client.extract_table(
            source=f"{source_project}.{source_dataset}.{source_table}",
            destination_uris=[destination_uri],
            job_config=job_config,
        )

        extract_job.result()  # Waits for the job to complete.

        # Verify that job was successful
        if extract_job.state != "DONE":
            log.error("Export failed with errors:", extract_job.errors)
            error_counter += 1

        # Initialize the storage client
        storage_client = storage.Client()
        bucket = storage_client.bucket(destination_bucket)
        blob = bucket.blob(f"{destination_prefix}/{temp_file}")

        # Read the temporary JSON file from GCS
        temp_file_content = blob.download_as_text()

        # Convert the content to a JSON array
        json_array = [json.loads(line) for line in temp_file_content.splitlines()]
        json_data = json.dumps(json_array, indent=1)

        # Write the JSON array to the final destination files in GCS:
        # 1. latest.json is a single file, that's easy to reference from Merino.
        # 2. {timestamp}.json keeps a historical record for debugging purposes.
        for suffix in ["latest", timestamp]:
            final_destination_uri = f"{destination_prefix}/{suffix}.json"
            final_blob = bucket.blob(final_destination_uri)
            final_blob.upload_from_string(json_data, content_type="application/json")

        # Delete the temporary file from GCS
        blob.delete()

        # Delete files older than 3 days
        delete_old_files(bucket, destination_prefix, deletion_days_old)

        log.info("Export successful and temporary file deleted")

    except Exception as err:
        error_counter += 1
        log.error(f"An error occurred: {err}")

        if error_counter > threshold:
            raise Exception(
                f"More than the accepted threshold of {threshold} operations failed."
            )


def delete_old_files(bucket, prefix, days_old):
    """Delete files older than `days_old` days from the bucket with the given prefix."""
    cutoff_date = datetime.now(timezone.utc) - timedelta(days=days_old)
    blobs = bucket.list_blobs(prefix=prefix)

    for blob in blobs:
        if blob.updated < cutoff_date:
            blob.delete()
            log.info(f"Deleted {blob.name}")



"""bigquery-etl CLI format command."""

import rich_click as click

from bigquery_etl.cli.utils import parallelism_option
from bigquery_etl.format_sql.format import format as format_sql


@click.command(
    help="""Format SQL files.

    Examples:

    # Format a specific file
    ./bqetl format sql/moz-fx-data-shared-prod/telemetry/core/view.sql

    # Format all SQL files in `sql/`
    ./bqetl format sql

    # Format standard in (will write to standard out)
    echo 'SELECT 1,2,3' | ./bqetl format
    """,
)
@click.argument(
    "paths",
    nargs=-1,
    type=click.Path(file_okay=True),
)
@click.option(
    "--check",
    default=False,
    is_flag=True,
    help="do not write changes, just return status;"
    " return code 0 indicates nothing would change;"
    " return code 1 indicates some files would be reformatted",
)
@parallelism_option()
def format(paths, check, parallelism):
    """Apply formatting to SQL files."""
    format_sql(paths, check=check, parallelism=parallelism)



"""PyTest plugin for collecting mypy tests on python scripts."""

from pytest_mypy import MypyFile, mypy_argv

from . import is_python_executable


def pytest_configure(config):
    """Add flag when running mypy.

    Fixes https://github.com/python/mypy/issues/1380
    """
    mypy_argv.append("--scripts-are-modules")


# adapted from
# https://github.com/dbader/pytest-mypy/blob/v0.7.0/src/pytest_mypy.py#L100-L111
def pytest_collect_file(parent, path):
    """Collect mypy tests."""
    if any(
        [parent.config.option.mypy, parent.config.option.mypy_ignore_missing_imports]
    ) and is_python_executable(path):
        return MypyFile.from_parent(parent=parent, fspath=path)



"""PyTest plugin for collecting black tests on python scripts."""

from pytest_black import BlackItem

from . import is_python_executable


# adapted from
# https://github.com/shopkeep/pytest-black/blob/0.3.10/pytest_black.py#L23-L29
def pytest_collect_file(parent, path):
    """Collect black tests."""
    config = parent.config
    if config.option.black and is_python_executable(path):
        return BlackItem.from_parent(parent, fspath=path)



import os

import pytest
import yaml
from click.testing import CliRunner

from bigquery_etl.cli.routine import create, info, rename


class TestRoutine:
    @pytest.fixture
    def runner(self):
        return CliRunner()

    def test_create_invalid_path(self, runner):
        with runner.isolated_filesystem():
            with open("foo.txt", "w") as f:
                f.write("")
            result = runner.invoke(create, ["udf.test_udf", "--path=foo.txt"])
            assert result.exit_code == 2

    def test_create_invalid_name(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/udf")
            result = runner.invoke(
                create,
                ["udf.udf.test_udf", "--udf"],
                obj={"DEFAULT_PROJECT": "moz-fx-data-shared-prod"},
            )
            assert (
                "New routine must be named like: <dataset>.<routine_name>"
                in result.output
            )
            assert result.exit_code == 1

    def test_create_missing_udf_procedure(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/udf")
            result = runner.invoke(
                create,
                ["udf.test_udf"],
                obj={"DEFAULT_PROJECT": "moz-fx-data-shared-prod"},
            )
            assert (
                "Please specify if new routine is a UDF or stored procedure"
                in result.output
            )
            assert result.exit_code == 1

    def test_create_udf(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/udf")
            result = runner.invoke(
                create,
                ["udf.test_udf", "--udf"],
                obj={"DEFAULT_PROJECT": "moz-fx-data-shared-prod"},
            )
            assert result.exit_code == 0
            assert os.listdir("sql/moz-fx-data-shared-prod/udf") == ["test_udf"]
            assert "udf.sql" in os.listdir("sql/moz-fx-data-shared-prod/udf/test_udf")
            assert "metadata.yaml" in os.listdir(
                "sql/moz-fx-data-shared-prod/udf/test_udf"
            )

    def test_create_stored_procedure(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/proc")
            result = runner.invoke(
                create,
                ["proc.test", "--stored_procedure"],
                obj={"DEFAULT_PROJECT": "moz-fx-data-shared-prod"},
            )
            assert result.exit_code == 0
            assert os.listdir("sql/moz-fx-data-shared-prod/proc") == ["test"]
            assert "stored_procedure.sql" in os.listdir(
                "sql/moz-fx-data-shared-prod/proc/test"
            )
            assert "metadata.yaml" in os.listdir(
                "sql/moz-fx-data-shared-prod/proc/test"
            )

    def test_create_mozfun_udf(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod")
            os.makedirs("sql/mozfun")
            result = runner.invoke(
                create,
                ["test_dataset.test_udf", "--udf"],
                obj={"DEFAULT_PROJECT": "mozfun"},
            )
            assert result.exit_code == 0
            assert os.listdir("sql/mozfun/test_dataset") == ["test_udf"]
            assert "udf.sql" in os.listdir("sql/mozfun/test_dataset/test_udf")
            assert "metadata.yaml" in os.listdir("sql/mozfun/test_dataset/test_udf")

    def test_routine_info(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/udf/test_udf")
            with open("sql/moz-fx-data-shared-prod/udf/test_udf/udf.sql", "w") as f:
                f.write("CREATE OR REPLACE FUNCTION udf.test_udf() AS (TRUE)")

            result = runner.invoke(
                info,
                ["udf.test_udf"],
                obj={"DEFAULT_PROJECT": "moz-fx-data-shared-prod"},
            )
            assert result.exit_code == 0
            assert "No metadata" in result.output
            assert "path:" in result.output

            metadata_conf = {"friendly_name": "test", "description": "test"}

            with open(
                "sql/moz-fx-data-shared-prod/udf/test_udf/metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(metadata_conf))

            result = runner.invoke(
                info,
                ["udf.test_udf"],
                obj={"DEFAULT_PROJECT": "moz-fx-data-shared-prod"},
            )
            assert result.exit_code == 0
            assert "No metadata" not in result.output
            assert "description" in result.output

    def test_udf_info_name_pattern(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/udf/test_udf")
            with open("sql/moz-fx-data-shared-prod/udf/test_udf/udf.sql", "w") as f:
                f.write("CREATE OR REPLACE FUNCTION udf.test_udf() AS (TRUE)")

            os.mkdir("sql/moz-fx-data-shared-prod/udf/another_udf")
            with open("sql/moz-fx-data-shared-prod/udf/another_udf/udf.sql", "w") as f:
                f.write("CREATE OR REPLACE FUNCTION udf.test_udf() AS (TRUE)")

            result = runner.invoke(
                info, ["udf.*"], obj={"DEFAULT_PROJECT": "moz-fx-data-shared-prod"}
            )
            assert result.exit_code == 0
            assert "udf.another_udf" in result.output
            assert "udf.test_udf" in result.output

            result = runner.invoke(
                info,
                ["udf.another*"],
                obj={"DEFAULT_PROJECT": "moz-fx-data-shared-prod"},
            )
            assert result.exit_code == 0
            assert "udf.another_udf" in result.output
            assert "udf.test_udf" not in result.output

    def test_udf_renaming_invalid_naming(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/udf/test_udf")
            with open("sql/moz-fx-data-shared-prod/udf/test_udf/udf.sql", "w") as f:
                f.write("CREATE OR REPLACE FUNCTION udf.test_udf() AS (TRUE)")

            os.mkdir("sql/moz-fx-data-shared-prod/udf/another_udf")
            with open("sql/moz-fx-data-shared-prod/udf/test_udf/udf.sql", "w") as f:
                f.write(
                    "CREATE OR REPLACE FUNCTION udf.another_udf() AS (udf.test_udf())"
                )

            result = runner.invoke(
                rename,
                ["udf.*", "something.else"],
                obj={"DEFAULT_PROJECT": "moz-fx-data-shared-prod"},
            )
            assert result.exit_code == 0

            result = runner.invoke(rename, ["dataset.udf"])
            assert result.exit_code == 2

    def test_udf_renaming(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/udf/test_udf")
            with open("sql/moz-fx-data-shared-prod/udf/test_udf/udf.sql", "w") as f:
                f.write("CREATE OR REPLACE FUNCTION udf.test_udf() AS (TRUE)")

            os.mkdir("sql/moz-fx-data-shared-prod/udf/another_udf")
            with open("sql/moz-fx-data-shared-prod/udf/another_udf/udf.sql", "w") as f:
                f.write(
                    "CREATE OR REPLACE FUNCTION udf.another_udf() AS (udf.test_udf())"
                )

            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT udf.test_udf()")

            result = runner.invoke(
                rename,
                ["udf.test_udf", "udf.renamed_udf"],
                obj={"DEFAULT_PROJECT": "moz-fx-data-shared-prod"},
            )
            assert result.exit_code == 0
            assert "test_udf" not in os.listdir("sql/moz-fx-data-shared-prod/udf")
            assert "another_udf" in os.listdir("sql/moz-fx-data-shared-prod/udf")
            assert "renamed_udf" in os.listdir("sql/moz-fx-data-shared-prod/udf")

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "r"
            ) as f:
                sql = f.read()
                assert "udf.renamed_udf" in sql
                assert "udf.test_udf" not in sql

            with open("sql/moz-fx-data-shared-prod/udf/another_udf/udf.sql", "r") as f:
                sql = f.read()
                assert "udf.renamed_udf" in sql
                assert "udf.test_udf" not in sql

            with open("sql/moz-fx-data-shared-prod/udf/renamed_udf/udf.sql", "r") as f:
                sql = f.read()
                assert "udf.renamed_udf" in sql
                assert "udf.test_udf" not in sql

    def test_mozfun_dataset_renaming(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/mozfun/udf/test_udf")
            with open("sql/mozfun/udf/test_udf/udf.sql", "w") as f:
                f.write("CREATE OR REPLACE FUNCTION udf.test_udf() AS (TRUE)")

            os.mkdir("sql/mozfun/udf/another_udf")
            with open("sql/mozfun/udf/another_udf/udf.sql", "w") as f:
                f.write(
                    "CREATE OR REPLACE FUNCTION udf.another_udf() AS (udf.test_udf())"
                )

            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT udf.test_udf()")

            result = runner.invoke(
                rename, ["udf.*", "new_dataset"], obj={"DEFAULT_PROJECT": "mozfun"}
            )
            assert result.exit_code == 0
            assert "udf" not in os.listdir("sql/mozfun")
            assert "new_dataset" in os.listdir("sql/mozfun")
            assert "another_udf" in os.listdir("sql/mozfun/new_dataset")
            assert "test_udf" in os.listdir("sql/mozfun/new_dataset")

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "r"
            ) as f:
                sql = f.read()
                assert "new_dataset.test_udf" in sql
                assert "udf.test_udf" not in sql

            with open("sql/mozfun/new_dataset/another_udf/udf.sql", "r") as f:
                sql = f.read()
                assert "new_dataset.test_udf" in sql
                assert "udf.test_udf" not in sql

            with open("sql/mozfun/new_dataset/test_udf/udf.sql", "r") as f:
                sql = f.read()
                assert "new_dataset.test_udf" in sql
                assert "udf.test_udf" not in sql



"""Experiment data downloaded from APIs, clean and upload to BigQuery."""

import datetime
import logging
from pathlib import Path

import click
import requests
import yaml
from google.cloud import bigquery

API_BASE_URL_EXPERIMENTS = "https://experimenter.services.mozilla.com"
API_BASE_URL_METRIC_HUB = (
    "https://api.github.com/repos/mozilla/metric-hub/contents/jetstream"
)

DEFAULT_PROJECT_ID = Path(__file__).parent.parent.parent.name
DEFAULT_DATASET_ID = Path(__file__).parent.parent.name
DEFAULT_TABLE_NAME = Path(__file__).parent.name

SCHEMA_FILE = Path(__file__).parent / "schema.yaml"
SCHEMA = bigquery.SchemaField.from_api_repr(
    {"name": "root", "type": "RECORD", **yaml.safe_load(SCHEMA_FILE.read_text())}
).fields


def parse_unix_datetime_to_string(unix_string):
    """Parse unix_string with milliseconds to date string."""
    if not unix_string:
        return None
    return datetime.datetime.fromtimestamp(int(unix_string) // 1000).strftime(
        "%Y-%m-%d"
    )


def get_api_response(url):
    """Get json of response if the requests to the API was successful."""
    response = requests.get(url)
    response.raise_for_status()
    return response.json()


def store_data_in_bigquery(data, schema, destination_project, destination_table_id):
    """Upload data to Bigquery in a single, non partitioned table."""
    client = bigquery.Client(project=destination_project)

    job_config = bigquery.LoadJobConfig(
        create_disposition="CREATE_IF_NEEDED",
        schema=schema,
        source_format=bigquery.SourceFormat.NEWLINE_DELIMITED_JSON,
        write_disposition="WRITE_TRUNCATE",
    )

    load_job = client.load_table_from_json(
        data, destination_table_id, location="US", job_config=job_config
    )
    load_job.result()
    stored_table = client.get_table(destination_table_id)
    logging.info(f"Loaded {stored_table.num_rows} rows into {destination_table_id}.")


def download_experiments_v1(url):
    """Download experiment data from API v1 and parse it."""
    experiments_v1 = []
    experiments = get_api_response(f"{url}/api/v1/experiments/")
    for experiment in experiments:
        if experiment["status"] == "Draft":
            continue
        experiments_v1.append(
            {
                "slug": experiment["slug"],
                "start_date": parse_unix_datetime_to_string(experiment["start_date"]),
                "enrollment_end_date": None,
                "end_date": parse_unix_datetime_to_string(experiment["end_date"]),
            }
        )
    logging.info(
        f"Downloaded {len(experiments_v1)} records from the experiments v1 API"
    )
    return experiments_v1


def download_experiments_v6(url):
    """Download experiment data from API v6 and parse it."""
    experiments_v6 = []
    experiments = get_api_response(f"{url}/api/v6/experiments/")
    for experiment in experiments:
        experiments_v6.append(
            {
                "slug": experiment["slug"],
                "start_date": experiment["startDate"],
                "enrollment_end_date": experiment["enrollmentEndDate"],
                "end_date": experiment["endDate"],
            }
        )
    logging.info(
        f"Downloaded {len(experiments_v6)} records from the experiments v6 API"
    )
    return experiments_v6


def download_metric_hub_files(url):
    """Download metric hub files from github."""
    metric_files = {}
    files = get_api_response(url)
    for file in files:
        if file["type"] != "file":
            continue
        slug = file["name"].removesuffix(".toml")
        metric_files[slug] = True
    logging.info(f"Downloaded {len(metric_files)} records from the API {url}")
    return metric_files


def compare_experiments_with_metric_hub_configs():
    """Download experiments from v1 and v6 API and compare them with config files in metric_hub."""
    experiments_v1 = download_experiments_v1(API_BASE_URL_EXPERIMENTS)
    experiments_v6 = download_experiments_v6(API_BASE_URL_EXPERIMENTS)
    metric_files = download_metric_hub_files(API_BASE_URL_METRIC_HUB)

    experiments = [
        (
            {**experiment, "has_config": True}
            if experiment["slug"] in metric_files
            else {**experiment, "has_config": False}
        )
        for experiment in (experiments_v1 + experiments_v6)
    ]
    return experiments


@click.command
@click.option(
    "--bq_project_id",
    default=DEFAULT_PROJECT_ID,
    show_default=True,
    help="BigQuery project the data is written to.",
)
@click.option(
    "--bq_dataset_id",
    default=DEFAULT_DATASET_ID,
    show_default=True,
    help="BigQuery dataset the data is written to.",
)
@click.option(
    "--bq_table_name",
    default=DEFAULT_TABLE_NAME,
    show_default=True,
    help="Bigquery table the data is written to.",
)
def run_experiments_stats(bq_project_id, bq_dataset_id, bq_table_name):
    """Download experiments from the APIs and store it in BigQuery."""
    experiments = compare_experiments_with_metric_hub_configs()
    destination_table_id = f"{bq_project_id}.{bq_dataset_id}.{bq_table_name}"
    store_data_in_bigquery(
        data=experiments,
        schema=SCHEMA,
        destination_project=bq_project_id,
        destination_table_id=destination_table_id,
    )


if __name__ == "__main__":
    logging.getLogger().setLevel(logging.INFO)
    run_experiments_stats()



"""Generate and run baseline_clients_last_seen queries for Glean apps."""

from sql_generators.glean_usage.common import GleanTable

TARGET_TABLE_ID = "baseline_clients_last_seen_v1"
PREFIX = "last_seen"
USAGE_TYPES = (
    "seen",
    "active",
    "created_profile",
    "seen_session_start",
    "seen_session_end",
)


class BaselineClientsLastSeenTable(GleanTable):
    """Represents generated baseline_clients_last_seen table."""

    def __init__(self):
        """Initialize baseline_clients_last_seen table."""
        GleanTable.__init__(self)
        self.target_table_id = TARGET_TABLE_ID
        self.prefix = PREFIX
        self.custom_render_kwargs = dict(
            usage_types=USAGE_TYPES,
        )



#!/usr/bin/env python3

"""Determine stable and derived table size partitions by performing dry runs."""

from argparse import ArgumentParser
from fnmatch import fnmatchcase
from functools import partial
from multiprocessing.pool import ThreadPool

from google.cloud import bigquery

parser = ArgumentParser(description=__doc__)
parser.add_argument("--date", required=True)  # expect string with format yyyy-mm-dd
parser.add_argument("--project", default="moz-fx-data-shared-prod")
parser.add_argument("--dataset", default=("*_derived", "*_stable"))  # pattern
parser.add_argument("--destination_dataset", default="monitoring_derived")
parser.add_argument("--destination_table", default="stable_and_derived_table_sizes_v1")


def get_tables(client, project, dataset):
    """Returns list of all available tables."""
    sql = f"SELECT dataset_id, table_id FROM `{project}.{dataset}.__TABLES__`"

    try:
        result = client.query(sql).result()
        return [(row.dataset_id, row.table_id) for row in result]
    except Exception as e:
        print(f"Error querying dataset {dataset}: {e}")

    return []


def get_partition_size_json(client, date, table):
    """Returns the size of a specific date parition of the specified table."""
    job_config = bigquery.QueryJobConfig(dry_run=True, use_query_cache=False)
    job_config_columns_info = bigquery.QueryJobConfig(
        dry_run=False, use_query_cache=False
    )
    dataset_id = table[0]
    table_id = table[1]

    partition_column_sql = f"""
                            SELECT column_name
                            FROM {dataset_id}.INFORMATION_SCHEMA.COLUMNS
                            WHERE table_name = '{table_id}' and is_partitioning_column = 'YES'
                        """
    partition_column_name_result = client.query(
        partition_column_sql, job_config=job_config_columns_info
    )

    partition_column_name = [row[0] for row in partition_column_name_result.result()]

    if len(partition_column_name) > 0 and partition_column_name[0] in (
        "submission_date",
        "submission_timestamp",
    ):
        sql = f"""
                SELECT * FROM `{dataset_id}.{table_id}`
                WHERE DATE({partition_column_name[0]}) = '{date}'
            """
        job = client.query(sql, job_config=job_config)
        size = job.total_bytes_processed if job.total_bytes_processed is not None else 0
        return {
            "submission_date": date,
            "dataset_id": dataset_id,
            "table_id": table_id,
            "byte_size": size,
        }


def save_table_sizes(client, table_sizes, date, destination_dataset, destination_table):
    """Writes paritions sizes for tables for a specific day to BigQuery."""

    job_config = bigquery.LoadJobConfig()
    job_config.schema = (
        bigquery.SchemaField("submission_date", "DATE"),
        bigquery.SchemaField("dataset_id", "STRING"),
        bigquery.SchemaField("table_id", "STRING"),
        bigquery.SchemaField("byte_size", "INT64"),
    )
    job_config.write_disposition = bigquery.job.WriteDisposition.WRITE_TRUNCATE

    partition_date = date.replace("-", "")
    client.load_table_from_json(
        table_sizes,
        f"{destination_dataset}.{destination_table}${partition_date}",
        job_config=job_config,
    ).result()


def main():
    args = parser.parse_args()
    client = bigquery.Client(args.project)
    stable_derived_partition_sizes = []

    for arg_dataset in args.dataset:
        stable_datasets = [
            dataset.dataset_id
            for dataset in list(client.list_datasets())
            if fnmatchcase(dataset.dataset_id, arg_dataset)
            and not fnmatchcase(dataset.dataset_id, "monitoring_derived")
        ]
        with ThreadPool(20) as p:
            stable_tables = p.map(
                partial(get_tables, client, args.project),
                stable_datasets,
                chunksize=1,
            )
            stable_tables = [table for tables in stable_tables for table in tables]

            partition_sizes = p.map(
                partial(get_partition_size_json, client, args.date),
                stable_tables,
                chunksize=1,
            )
            partition_sizes = filter(lambda x: x is not None, partition_sizes)
            stable_derived_partition_sizes.extend(partition_sizes)

    save_table_sizes(
        client,
        stable_derived_partition_sizes,
        args.date,
        args.destination_dataset,
        args.destination_table,
    )


if __name__ == "__main__":
    main()
